/* 

   ICFP 2001 Contest entry
   Chris Eagle
   
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#define b_tag 0x80000000
#define em_tag 0x40000000
#define i_tag 0x20000000
#define s_tag 0x10000000
#define u_mask 0xC0000000
#define u_count 0x40000000
#define size_tag 0x0F
#define r_color 0x2000000
#define g_color 0x1000000
#define b_color 0x800000
#define c_color 0x400000
#define m_color 0x200000
#define y_color 0x100000
#define k_color 0x80000
#define w_color 0x40000
#define tt_flag 0x20000
#define color_mask 0x03FC0000
#define plain_mask 0x03FDFFFF

typedef struct runTag {
   unsigned int attr;
   int tag;
   void *child;
   struct runTag *next;
   struct runtag *parent;
} Run;

typedef struct stackTag {
   struct runTag *open;
   struct stackTag *next;
} Stack;

Run root = {0, 1, 0, 0, 0};
Run *active = &root;

Stack *sp = 0;

unsigned int attr = 0;
unsigned int text = 0;

Run *pop() {
   Stack *temp = sp;
   Run *r = sp->open;
   sp = sp->next;
   free(temp);
   return r;
}

void push(Run *r) {
   Stack *node;
   node = (Stack*) malloc(sizeof(Stack));
   node->next = sp;
   node->open = r;
   sp = node;
}

int addChild(Run *parent, Run *child) {
   child->parent = parent;
   if (parent->child) {
      parent = (Run*) parent->child;
      while (parent->next) parent = parent->next;
      if (child->tag == parent->tag) {
         //this merges things like <B>x</B><B>y</B> into <B>xy</B>
         //<EM>xyz</EM><EM>abc</EM> should go to <EM>xyzabc</EM> as well
         if (child->tag != 'U' && child->tag != -1) {
            active = parent;
            push(parent);
            free(child);
            return 0;  //merged adjacent nodes, keep using same parent
         }
         else if (child->tag == 'U') {
            if ((child->attr & u_mask) == (parent->attr & u_mask)) {
               active = parent;
               push(parent);
               free(child);
               return 0;  //merged adjacent nodes, keep using same parent
            }
         }
      }
      parent->next = child;
   }
   else {
      parent->child = child;
   }
   return 1;  //a new parent capable node is in effect
}

void activate(Run *r) {
   if (addChild(active, r)) {
      active = r;
      push(r);
   }
}

void optimize() {
}

void closeTag() {   //***eliminate childless nodes here
   Run * r = pop();
   Run *temp;
   if (r) {  //pop back to parent if we weren't eating this node
      if (!active->child) {
         temp = active->parent;
         if (temp->child == active) {
            temp->child = NULL;
            free(active);
            active = temp;
         }
         else {
            temp = temp->child;
            while (temp->next != active) {
               temp = temp->next;
            }
            temp->next = NULL;
            temp = active;
            active = active->parent;
            free(temp);
         }            
      }
      else {
         active = active->parent;
      }
      attr = active->attr;
   }
}

void openTag(unsigned int tag) {
   Run *child = (Run *) calloc(1, sizeof(Run));
   child->tag = tag;
   child->attr = attr;
   if (tag) {  //don't activate absorbed nodes, the parent will handle
      activate(child);
   }
   else {
      push(NULL); //place holder absorbed nodes, causes closing tag to get eaten
   }
}

void newText(int ch) {
   Run *child = (Run *) calloc(1, sizeof(Run));
   child->tag = -1;
   child->attr = 4;
   activate(child);
   child->child = malloc(4);
   ((char*)(child->child))[0] = ch;
   ((char*)(child->child))[1] = 0;
   text = 1;
}

void append(int ch) {
   char *p1 = (char *)active->child;
   unsigned int len = 0;
   while (p1[len]) len++;
   if (len == (active->attr - 1)) {
      p1 = active->child = realloc(active->child, 2 * active->attr);
      active->attr *= 2;
   }
   p1[len++] = ch;
   p1[len] = 0;
}

void tag() {
   int ch;
   ch = getchar();
   if (ch == '/') {
      getchar();
      closeTag();
      while (getchar() != '>');
      return;
   }
   switch (ch) {
      case 'B':
         if (attr & b_tag) ch = 0;  //absorb B's within B's
         else attr |= b_tag;
         break;
      case 'E':
         getchar();  //should be 'M'
         if (!(attr & s_tag)) attr ^= em_tag;
         else ch = 0;  //absorb EM's within S's
         break;
      case 'I':
         if (attr & i_tag) ch = 0; //absorb I's within I's
         else attr |= i_tag;
         break;
      case 'P':
         getchar();  //should be 'L'
         attr &= plain_mask;
         break;
      case 'S':
         attr = (attr | s_tag) & ~em_tag;
         break;
      case 'T':
         getchar();  //should be 'T'
         if (attr & tt_flag) ch = 0; //absorb TT's within TT's
         else attr |= tt_flag;
         break;
      case 'U':
         if ((attr & u_mask) == u_mask) ch = 0; //absorb U3's within U3's
         else attr += u_count;
         break;
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
         if ((attr & size_tag) == (ch & size_tag)) ch = 0; //absorb n's within n's
         else attr = (attr & ~size_tag) | (ch & size_tag);
         break;
      case 'r':
         if (attr & r_color) ch = 0;  //absorb r's within r's
         else attr = (attr & ~color_mask) | r_color;
         break;
      case 'g':
         if (attr & g_color) ch = 0;  //absorb g's within g's
         else attr = (attr & ~color_mask) | g_color;
         break;
      case 'b':
         if (attr & b_color) ch = 0;  //absorb b's within b's
         else attr = (attr & ~color_mask) | b_color;
         break;
      case 'c':
         if (attr & c_color) ch = 0;  //absorb c's within c's
         else attr = (attr & ~color_mask) | c_color;
         break;
      case 'm':
         if (attr & m_color) ch = 0;  //absorb m's within m's
         else attr = (attr & ~color_mask) | m_color;
         break;
      case 'y':
         if (attr & y_color) ch = 0;  //absorb y's within y's
         else attr = (attr & ~color_mask) | y_color;
         break;
      case 'k':
         if (attr & k_color) ch = 0;  //absorb k's within k's
         else attr = (attr & ~color_mask) | k_color;
         break;
      case 'w':
         if (attr & w_color) ch = 0;  //absorb w's within w's
         else attr = (attr & ~color_mask) | w_color;
         break;
   }
   openTag(ch);
   getchar();  //should be a '>'
}

void dumpTree(Run *ptr) {
   if (!ptr) return;
   if (ptr->tag == -1) {  //this is a CDATA node
      fputs((char*) ptr->child, stdout);
      dumpTree(ptr->next);
      return;
   }
   if (ptr->child) { //dont print nodes w/ no children
      //this node eliminates the double negative oe <EM><EM>xyz</EM></EM>
      if (ptr->tag == 'E' && ((Run*)ptr->child)->tag == 'E' && !((Run*)ptr->child)->next) {
         dumpTree((Run*) ((Run*)ptr->child)->child);
      }
      else {
         if (ptr->tag > 1) {
            fputc('<', stdout);
            fputc(ptr->tag, stdout);
            switch (ptr->tag) {
               case 'E':
                  fputc('M', stdout);
                  break;
               case 'P':
                  fputc('L', stdout);
                  break;
               case 'T':
                  fputc('T', stdout);
                  break;
            }
            fputc('>', stdout);
         }
         dumpTree((Run*) ptr->child);
         if (ptr->tag > 1) {
            fputc('<', stdout);
            fputc('/', stdout);
            fputc(ptr->tag, stdout);
            switch (ptr->tag) {
               case 'E':
                  fputc('M', stdout);
                  break;
               case 'P':
                  fputc('L', stdout);
                  break;
               case 'T':
                  fputc('T', stdout);
                  break;
            }
            fputc('>', stdout);
         }
      }
   }
   dumpTree(ptr->next);
}

void parse() {
   int ch;

   while ((ch = getchar()) != EOF) {
      if (!(attr & tt_flag) && isspace(ch)) {
         while (isspace(ch = getchar()));
         ungetc(ch, stdin);
         ch = ' ';
      }
      switch (ch) {
         case '<':
            if (text) {
               closeTag();
            }
            text = 0;
            tag();
            break;
         default:
            if (text) {
               append(ch);
            }
            else {
               newText(ch);
            }
            break;
      }
   }
   
   optimize();
   
   dumpTree(&root);
}



int main() {

   parse();

   return 0;
}

